;
;                    Inertia Realtime 3D Rendering Engine
;    Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;

;
; PERSPASM.INC
; code: Alex Chalfin
;

; Perspective based polygon include file

; paramters:
; edi  - Polygon pointer value

  Mov  eax,[edi].PolyAPtr       ; VA = P->A
  Mov  ebx,[edi].PolyBPtr       ; VB = P->B
  Mov  ecx,[edi].PolyCPtr       ; VC = P->C
                                ;
                                ;
IFDEF  __INTERP_UV__            ;
  Lea  esi,[edi].UV1x           ;  TVA = &P->UV1
  Mov  [TVA],esi                ;  Store the address of first texture coord
                                ;
  Lea  esi,[edi].UV2x           ;  TVB = &P->UV2
  Mov  [TVB],esi                ;  Store the address of second texture coord
                                ;
  Lea  esi,[edi].UV3x           ;  TVC = &P->UV3
  Mov  [TVC],esi                ;  Store the address of third texture coord
ENDIF                           ;
                                ;
;---------------------------------------------------------------------------
; First bubble sort compare
;---------------------------------------------------------------------------
                                ;
  fld  dword ptr [eax].ScrY     ;
  fld  dword ptr [ebx].ScrY     ;
  fcompp                        ; if (VA->Scr.y > VB->Scr.y)
                                ;
  Push  eax                     ; save first vertex pointer
  fstsw ax                      ;
  sahf                          ; store FPU flags
  Pop   eax                     ;
                                ;
  ja  @@SkipPointerSwap1        ; if st(0) <= st(1) then jump
                                ;
  Xchg eax,ebx                  ; Swap the A and B pointers
                                ;
IFDEF __INTERP_UV__             ;
  Mov  edx,[TVA]                ;
  Xchg edx,[TVB]                ;
  Mov  [TVA],edx                ;
ENDIF                           ;
                                ;
 @@SkipPointerSwap1:            ;
                                ;
;---------------------------------------------------------------------------
; Second bubble sort compare
;---------------------------------------------------------------------------
                                ;
  fld  dword ptr [ebx].ScrY     ;
  fld  dword ptr [ecx].ScrY     ;
  fcompp                        ; if (VB->Scr.y > VC->Scr.y)
                                ;
  Push  eax                     ; Save first vertex pointer
  fstsw ax                      ;
  sahf                          ; store FPU flags
  Pop   eax                     ;
                                ;
  ja  @@SkipPointerSwap2        ; if st(0) <= st(1) then jump
                                ;
  Xchg ebx,ecx                  ; Swap the B and C pointers
                                ;
IFDEF __INTERP_UV__             ;
  Mov  edx,[TVB]                ;
  Xchg edx,[TVC]                ;
  Mov  [TVB],edx                ;
ENDIF                           ;
                                ;
 @@SkipPointerSwap2:            ;
                                ;
;---------------------------------------------------------------------------
; Third bubble sort compare
;---------------------------------------------------------------------------
                                ;
  fld  dword ptr [eax].ScrY     ;
  fld  dword ptr [ebx].ScrY     ;
  fcompp                        ; if (VA->Scr.y > VB->Scr.y)
                                ;
  Push  eax                     ; Save first vertex pointer
  fstsw ax                      ;
  sahf                          ; store FPU flags
  Pop   eax                     ;
                                ;
  ja  @@SkipPointerSwap3        ; if st(0) <= st(1) then jump
                                ;
  Xchg eax,ebx                  ; Swap the A and B pointers
                                ;
IFDEF __INTERP_UV__             ;
  Mov  edx,[TVA]                ;
  Xchg edx,[TVB]                ;
  Mov  [TVA],edx                ;
ENDIF                           ;
                                ;
 @@SkipPointerSwap3:            ;
                                ;
;----------------------------------------------------------------------------
; Store the vertex pointers
;----------------------------------------------------------------------------
                                ;
  Mov  [VertPtrA],eax           ;  Store the vertex pointers into
  Mov  [VertPtrB],ebx           ;  variables for later use
  Mov  [VertPtrC],ecx           ;
                                ;
;----------------------------------------------------------------------------
; Calculate gouraud intensities if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld   dword ptr [eax].VShade  ; load the first vertex shade intensity
  fmul  dword ptr [_tmp_flt_63] ; scale it up to yield 64 intensities
                                ;
  fld   dword ptr [ebx].VShade  ; second vertex shade value
  fmul  dword ptr [_tmp_flt_63] ;
                                ;
  fld   dword ptr [ecx].VShade  ; third vertex shade value
  fmul  dword ptr [_tmp_flt_63] ;
                                ;
  fstp  dword ptr [FLTInten3]   ; save the values
  fstp  dword ptr [FLTInten2]   ;
  fstp  dword ptr [FLTInten1]   ;
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate texture coords if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_UV__             ;
                                ;
  Push eax                      ; save the vertex pointers for later
  Push ebx                      ;
  Push ecx                      ;
  Push edi                      ;
                                ;
  Mov  edx,eax                  ; store vert A pointer
  Mov  edi,ebx                  ; store vert B pointer
  Mov  esi,ecx                  ; store vert C pointer
                                ;
  Mov  eax,[TVA]                ; grab the texture pointers
  Mov  ebx,[TVB]                ;
  Mov  ecx,[TVC]                ;
                                ;
  fld  dword ptr [eax].X2D      ; get the first texture X value
  fmul dword ptr [_tmp_flt_255] ; scale it up
  fmul dword ptr [edx].OneOverZ ; multiply by one over z
                                ;
  fld  dword ptr [eax].Y2D      ; get the first texture Y value
  fmul dword ptr [_tmp_flt_255] ; scale it up
  fmul dword ptr [edx].OneOverZ ; multiply by one over z
                                ;
                                ;
  fstp dword ptr [FLTTextureY1] ;  store it in the proper variables
  fstp dword ptr [FLTTextureX1] ;
                                ;
  fld  dword ptr [ebx].X2D      ;  get the second texture X value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
  fmul dword ptr [edi].OneOverZ ; multiply by one over z
                                ;
  fld  dword ptr [ebx].Y2D      ;  get the second texture Y value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
  fmul dword ptr [edi].OneOverZ ; multiply by one over z
                                ;
  fstp dword ptr [FLTTextureY2] ;  store it in the proper variables
  fstp dword ptr [FLTTextureX2] ;
                                ;
  fld  dword ptr [ecx].X2D      ;  get the third texture X value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
  fmul dword ptr [esi].OneOverZ ; multiply by one over z
                                ;
  fld  dword ptr [ecx].Y2D      ;  get the third texture Y value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
  fmul dword ptr [esi].OneOverZ ; multiply by one over z
                                ;
  fstp dword ptr [FLTTextureY3] ;  store it in the proper variables
  fstp dword ptr [FLTTextureX3] ;
                                ;
  Pop  edi                      ;
  Pop  ecx                      ; restore the vertex pointer
  Pop  ebx                      ;
  Pop  eax                      ;
                                ;
;----------------------------------------------------------------------------
; create one over Z variables
;----------------------------------------------------------------------------
                                ;
  Mov  esi,[eax].OneOverZ       ;
  Mov  edx,[ebx].OneOverZ       ;
                                ;
  Mov  [FLTZ1],esi              ; Store Z1
  Mov  [FLTZ2],edx              ; Store Z2
                                ;
  Mov  edx,[ecx].OneOverZ       ;
  Mov  [FLTZ3],edx              ; Store Z3
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate environment mapping  coords if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_ENV__            ;
                                ;
  fld  dword ptr [eax].EnvX     ;  get the first env X value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
  fld  dword ptr [eax].EnvY     ;  get the first env Y value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
                                ;
  fstp dword ptr [FLTEnvY1]     ;  store it in the proper variables
  fstp dword ptr [FLTEnvX1]     ;
                                ;
  fld  dword ptr [ebx].EnvX     ;  get the second env X value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
  fld  dword ptr [ebx].EnvY     ;  get the second env Y value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
                                ;
  fstp dword ptr [FLTEnvY2]     ;  store it in the proper variables
  fstp dword ptr [FLTEnvX2]     ;
                                ;
  fld  dword ptr [ecx].EnvX     ;  get the third Env X value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
  fld  dword ptr [ecx].EnvY     ;  get the third Env Y value
  fmul dword ptr [_tmp_flt_255] ;  scale it up
                                ;
  fstp dword ptr [FLTEnvY3]     ;  store it in the proper variables
  fstp dword ptr [FLTEnvX3]     ;
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate height parameters for the polygon
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [eax].ScrY     ;
  fadd dword ptr [_tmp_half]    ;
  fistp dword ptr [CeilY1]      ; ceilY1 = FLT2INT(ceil(VA->Scr.y));
                                ;
  fld  dword ptr [ebx].ScrY     ;
  fadd dword ptr [_tmp_half]    ;
  fistp dword ptr [CeilY2]      ; ceilY2 = FLT2INT(ceil(VB->Scr.y));
                                ;
  fld  dword ptr [ecx].ScrY     ;
  fadd dword ptr [_tmp_half]    ;
  fistp dword ptr [CeilY3]      ; ceilY3 = FLT2INT(ceil(VC->Scr.y));
                                ;
  Mov edx,[CeilY3]              ;
  Mov esi,[CeilY2]              ;
                                ;
  Sub edx,[CeilY1]              ; Calculate the Y size of the polygon
  Sub esi,[CeilY1]              ; calc the Y size of the top
                                ;
  Mov [TotalHeight],edx         ;
  Mov [TopHeight],esi           ;
                                ;
  Sub edx,esi                   ;
  Mov [BottomHeight],edx        ; calc the Y size of the bottom
                                ;
  Mov edx,[TotalHeight]         ;
  Cmp edx,0                     ;
  Jle @@ExitPolygonFiller       ;
                                ;
;----------------------------------------------------------------------------
; Calculate the delta for the long side of the polygon
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [ecx].ScrY     ;
  fsub dword ptr [eax].ScrY     ;
                                ;
  fld1                          ;
  fdivrp st(1),st(0)            ; OneOverTotalDY  = 1.0 / (VC->Scr.y - VA->Scr.y);
                                ;
  fst    [OneOverTotalDY]       ;
                                ;
  fld    dword ptr [ecx].ScrX   ;
  fsub   dword ptr [eax].ScrX   ;
  fmulp  st(1),st(0)            ; (VC->Scr.x - VA->Scr.x) * OneOverDY;
                                ;
  fst    dword ptr [FLTLongDX]  ; save floating value for sub-pixel correction
                                ;
  fld    st(0)                  ;
  fmul   dword ptr [_tmp_fixone];
  fistp  dword ptr [FXDLongDX]  ;  convert the delta value to fixed
                                ;
  fild   dword ptr [CeilY1]     ;
  fsub   dword ptr [eax].ScrY   ;
  fmulp  st(1),st(0)            ; (ceilY1 - VA->Scr.y) * t
  fadd   dword ptr [eax].ScrX   ;
  fmul   dword ptr [_tmp_fixone];
                                ;
  fistp  dword ptr [FXDLongX]   ; longx  = FLT2FXD(VA->Scr.x + (ceilY1 - VA->Scr.y) * t);
                                ;
;----------------------------------------------------------------------------
; Calculate x delta value for the polygon
;----------------------------------------------------------------------------
                                ;
                                ; calculate 2d cross product to determine
                                ; orientation of polygon
  fld    dword ptr [ecx].ScrX   ;
  fsub   dword ptr [eax].ScrX   ; (VC->Scr.x - VA->Scr.x)
  fld    dword ptr [ebx].ScrY   ;
  fsub   dword ptr [eax].ScrY   ; (VB->Scr.y - VA->Scr.y)
  fmulp  st(1),st(0)            ;
                                ;
  fld    dword ptr [ecx].ScrY   ;
  fsub   dword ptr [eax].ScrY   ; (VC->Scr.y - VA->Scr.y)
  fld    dword ptr [ebx].ScrX   ;
  fsub   dword ptr [eax].ScrX   ; (VB->Scr.x - VA->Scr.x)
  fmulp  st(1),st(0)            ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fld1                          ;
  fdivrp st(1),st(0)            ;
                                ;
  fstp   dword ptr [PolyDX]     ;
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal gouraud delta value for the polygon if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld   dword ptr [FLTInten3]   ;
  fsub  dword ptr [FLTInten1]   ; (c3 - c1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [FLTInten2]   ;
  fsub  dword ptr [FLTInten1]   ; (c2 - c1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [PolyDX]     ;
  fmul   dword ptr [FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fmul   dword ptr [_tmp_fixone];
  fistp  dword ptr [HorizontalDC];
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal texture delta values for the polygon if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_UV__             ;
                                ;
  fld   dword ptr [FLTTextureX3];
  fsub  dword ptr [FLTTextureX1]; (u3 - u1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [FLTTextureX2];
  fsub  dword ptr [FLTTextureX1]; (u2 - u1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [PolyDX]     ;
  fmul   dword ptr [FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fstp  dword ptr [HorizontalDU]; Keep in floating point
                                ;
  fld   dword ptr [FLTTextureY3];
  fsub  dword ptr [FLTTextureY1]; (v3 - v1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [FLTTextureY2];
  fsub  dword ptr [FLTTextureY1]; (v2 - v1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [PolyDX]     ;
  fmul   dword ptr [FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fstp  dword ptr [HorizontalDV]; Keep in floating point
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal 1/z delta
;----------------------------------------------------------------------------
                                ;
  fld   dword ptr [FLTZ3]       ;
  fsub  dword ptr [FLTZ1]       ; (z3 - z1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [FLTZ2]       ;
  fsub  dword ptr [FLTZ1]       ; (z2 - z1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [PolyDX]     ;
  fmul   dword ptr [FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fstp  dword ptr [HorizontalDZ]; Keep in floating point
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate horizontal env delta values for the polygon if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_ENV__            ;
                                ;
  fld   dword ptr [FLTEnvX3]    ;
  fsub  dword ptr [FLTEnvX1]    ; (s3 - s1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [FLTEnvX2]    ;
  fsub  dword ptr [FLTEnvX1]    ; (s2 - s1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [PolyDX]     ;
  fmul   dword ptr [FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fmul   dword ptr [_tmp_fixone];
  fistp  dword ptr [HorizontalDS];
                                ;
  fld   dword ptr [FLTEnvY3]    ;
  fsub  dword ptr [FLTEnvY1]    ; (t3 - t1)
  fld   dword ptr [ebx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VB->Scr.y - VA->Scr.y)
  fmulp st(1),st(0)             ;
                                ;
  fld   dword ptr [ecx].ScrY    ;
  fsub  dword ptr [eax].ScrY    ; (VC->Scr.y - VA->Scr.y)
  fld   dword ptr [FLTEnvY2]    ;
  fsub  dword ptr [FLTEnvY1]    ; (t2 - t1)
  fmulp st(1),st(0)             ;
                                ;
  fsubrp st(1),st(0)            ;
                                ;
  fmul   dword ptr [PolyDX]     ;
  fmul   dword ptr [FUDGE_FACTOR];  Sshhh! don't tell anyone this is here :)
                                ;
  fmul   dword ptr [_tmp_fixone];
  fistp  dword ptr [HorizontalDT];
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate OneOverTopDY
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[TopHeight]          ;
  Cmp  edx,0                    ;
  Jle @@SkipTopYHeightDelta     ;
                                ;
  fld  dword ptr [ebx].ScrY     ;
  fsub dword ptr [eax].ScrY     ;
                                ;
  fld1                          ;
  fdivrp  st(1),st(0)           ; 1.0 / (VB->Scr.y - VA->Scr.y)
  fstp  dword ptr [OneOverTopDY];
                                ;
 @@SkipTopYHeightDelta:         ;
                                ;
;----------------------------------------------------------------------------
; Calculate OneOverBotDY
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[BottomHeight]       ;
  Cmp  edx,0                    ;
  Jle @@SkipBotYHeightDelta     ;

  fld  dword ptr [ecx].ScrY     ;
  fsub dword ptr [ebx].ScrY     ;
                                ;
  fld1                          ;
  fdivrp  st(1),st(0)           ; 1.0 / (VC->Scr.y - VB->Scr.y)
  fstp  dword ptr [OneOverBotDY];
                                ;
 @@SkipBotYHeightDelta:         ;
                                ;
;----------------------------------------------------------------------------
; Calculate y-deltas. Done only when necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_EXTRA__          ;
                                ;
  fld  dword ptr [PolyDX]       ;
  fldz                          ;
  fcompp                        ; if (plydx < 0.0)
                                ;
  Push  eax                     ; Save first vertex pointer
  fstsw ax                      ;
  sahf                          ;
                                ;
  Pop   eax                     ;
                                ;
  jae @@ShortSideLeftDelta      ;
                                ;
;----------------------------------------------------------------------------
; Calculate long gouraud deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld  dword ptr [FLTInten3]    ;
  fsub dword ptr [FLTInten1]    ;
  fmul dword ptr [OneOverTotalDY];  (c3 - c1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1DC]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dc
  fadd  dword ptr [FLTInten1]   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1C]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long texture deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_UV__            ;
                                ;
  fld  dword ptr [FLTTextureX3] ;
  fsub dword ptr [FLTTextureX1] ;
  fmul dword ptr [OneOverTotalDY];  (u3 - u1) * OneOverDY
                                ;
  fst  dword ptr [FXDSide1DU]   ; keep as floating point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1du
  fadd  dword ptr [FLTTextureX1];
  fstp  dword ptr [FXDSide1U]   ; keep as floating point
                                ;
  fld  dword ptr [FLTTextureY3] ;
  fsub dword ptr [FLTTextureY1] ;
  fmul dword ptr [OneOverTotalDY];  (v3 - v1) * OneOverDY
                                ;
  fst  dword ptr [FXDSide1DV]   ; keep floating point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dv
  fadd  dword ptr [FLTTextureY1];
  fstp  dword ptr [FXDSide1V]   ; keep floating point
                                ;
                                ;
  fld  dword ptr [FLTZ3]        ;
  fsub dword ptr [FLTZ1]        ;
  fmul dword ptr [OneOverTotalDY];  (z3 - z1) * OneOverDY
                                ;
  fst  dword ptr [FXDSide1DZ]   ; keep floating point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dz
  fadd  dword ptr [FLTZ1]       ;
  fstp  dword ptr [FXDSide1Z]   ; keep floating point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long env deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_ENV__           ;
                                ;
  fld  dword ptr [FLTEnvX3]     ;
  fsub dword ptr [FLTEnvX1]     ;
  fmul dword ptr [OneOverTotalDY];  (s3 - s1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1DS]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1ds
  fadd  dword ptr [FLTEnvX1]    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1S]   ; convert to fixed point
                                ;
  fld  dword ptr [FLTEnvY3]     ;
  fsub dword ptr [FLTEnvY1]     ;
  fmul dword ptr [OneOverTotalDY];  (t3 - t1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1DT]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dt
  fadd  dword ptr [FLTEnvY1]    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1T]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Jump to the next section
;----------------------------------------------------------------------------
                                ;
  Jmp  @@BeginRendering         ; Jump to the actual rasterization process
 @@ShortSideLeftDelta:          ;
                                ;
;----------------------------------------------------------------------------
; Calculate top deltas if the top of the polygon exists
;----------------------------------------------------------------------------
                                ;
  Mov edx,[TopHeight]           ;
  Cmp edx,0                     ;
  Jle @@CalcBottomHeightDeltas  ; check to make sure there is a top
                                ;
;----------------------------------------------------------------------------
; Calculate top short gouraud deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld  dword ptr [FLTInten2]    ;
  fsub dword ptr [FLTInten1]    ;
  fmul dword ptr [OneOverTopDY] ;  (c2 - c1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1DC]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dc
  fadd  dword ptr [FLTInten1]   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1C]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate short top texture deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_UV__            ;
                                ;
  fld  dword ptr [FLTTextureX2] ;
  fsub dword ptr [FLTTextureX1] ;
  fmul dword ptr [OneOverTopDY] ;  (u2 - u1) * OneOverDY
                                ;
  fst   dword ptr [FXDSide1DU]  ; keep floating point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1du
  fadd  dword ptr [FLTTextureX1];
  fstp  dword ptr [FXDSide1U]   ; keep floating point
                                ;
  fld  dword ptr [FLTTextureY2] ;
  fsub dword ptr [FLTTextureY1] ;
  fmul dword ptr [OneOverTopDY] ;  (v2 - v1) * OneOverDY
                                ;
  fst   dword ptr [FXDSide1DV]  ; keep floating point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dv
  fadd  dword ptr [FLTTextureY1];
  fstp  dword ptr [FXDSide1V]   ; keep floating point
                                ;
  fld  dword ptr [FLTZ2]        ;
  fsub dword ptr [FLTZ1]        ;
  fmul dword ptr [OneOverTopDY] ;  (z2 - z1) * OneOverDY
                                ;
  fst   dword ptr [FXDSide1DZ]  ; keep floating point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dz
  fadd  dword ptr [FLTZ1]       ;
  fstp  dword ptr [FXDSide1Z]   ; keep floating point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long env deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_ENV__           ;
                                ;
  fld  dword ptr [FLTEnvX2]     ;
  fsub dword ptr [FLTEnvX1]     ;
  fmul dword ptr [OneOverTopDY] ;  (s2 - s1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1DS]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1ds
  fadd  dword ptr [FLTEnvX1]    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1S]   ; convert to fixed point
                                ;
  fld  dword ptr [FLTEnvY2]     ;
  fsub dword ptr [FLTEnvY1]     ;
  fmul dword ptr [OneOverTopDY] ;  (t3 - t1) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1DT]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1dt
  fadd  dword ptr [FLTEnvY1]    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide1T]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Check to see if the bottom half of the polygon exists
;----------------------------------------------------------------------------
                                ;
 @@CalcBottomHeightDeltas:      ;
                                ;
  Mov  edx,[BottomHeight]       ;
  Cmp  edx,0                    ;
  Jle  @@BeginRendering         ;
                                ;
;----------------------------------------------------------------------------
; Calculate top short gouraud deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF __INTERP_C__              ;
                                ;
  fld  dword ptr [FLTInten3]    ;
  fsub dword ptr [FLTInten2]    ;
  fmul dword ptr [OneOverBotDY] ;  (c3 - c2) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide2DC]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dc
  fadd  dword ptr [FLTInten2]   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide2C]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate short top texture deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_UV__            ;
                                ;
  fld  dword ptr [FLTTextureX3] ;
  fsub dword ptr [FLTTextureX2] ;
  fmul dword ptr [OneOverBotDY] ; (u3 - u2) * OneOverDY
                                ;
  fst  dword ptr [FXDSide2DU]   ; keep floating point
                                ;
  fild  dword ptr [CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2du
  fadd  dword ptr [FLTTextureX2];
  fstp  dword ptr [FXDSide2U]   ; convert to fixed point
                                ;
  fld  dword ptr [FLTTextureY3] ;
  fsub dword ptr [FLTTextureY2] ;
  fmul dword ptr [OneOverBotDY] ;  (v2 - v1) * OneOverDY
                                ;
  fst  dword ptr [FXDSide2DV]   ;  keep floating point
                                ;
  fild  dword ptr [CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dv
  fadd  dword ptr [FLTTextureY2];
  fstp  dword ptr [FXDSide2V]   ; keep floating point
                                ;
  fld  dword ptr [FLTZ3]        ;
  fsub dword ptr [FLTZ2]        ;
  fmul dword ptr [OneOverBotDY] ;  (z2 - z1) * OneOverDY
                                ;
  fst  dword ptr [FXDSide2DZ]   ;  keep floating point
                                ;
  fild  dword ptr [CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dv
  fadd  dword ptr [FLTZ2]       ;
  fstp  dword ptr [FXDSide2Z]   ; keep floating point
                                ;
ENDIF                           ;
                                ;
;----------------------------------------------------------------------------
; Calculate long env deltas if necessairy
;----------------------------------------------------------------------------
                                ;
IFDEF  __INTERP_ENV__           ;
                                ;
  fld  dword ptr [FLTEnvX3]     ;
  fsub dword ptr [FLTEnvX2]     ;
  fmul dword ptr [OneOverBotDY] ;  (s3 - s2) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide2DS]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY1 - VA->Scr.y)
  fmulp st(1),st(0)             ; (CeilY1 - VA->Scr.y) * fltside1ds
  fadd  dword ptr [FLTEnvX2]    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide2S]   ; convert to fixed point
                                ;
  fld  dword ptr [FLTEnvY3]     ;
  fsub dword ptr [FLTEnvY2]     ;
  fmul dword ptr [OneOverBotDY] ;  (t3 - t2) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide2DT]  ;  convert to fixed point
                                ;
  fild  dword ptr [CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ; (CeilY2 - VB->Scr.y)
  fmulp st(1),st(0)             ; (CeilY2 - VB->Scr.y) * fltside2dt
  fadd  dword ptr [FLTEnvY2]    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDSide2T]   ; convert to fixed point
                                ;
ENDIF                           ;
                                ;
ENDIF                           ;  __INTERP_EXTRA__
;----------------------------------------------------------------------------
; Lets do some rendering!!!
;----------------------------------------------------------------------------
                                ;
 @@BeginRendering:              ; the start of the rasterization process
                                ;
  Push  eax                     ; Save vertex A pointer
                                ;
  Mov   eax,[CeilY1]            ;
  Mov   edx,[LineSize]          ;
  IMul  edx                     ;
  Add   eax,[VirtualAddress]    ; add in the address of the virtual page
  Mov   [YAddress],eax          ;
                                ;
  Pop   eax                     ; restore vertex A pointer
                                ;
;----------------------------------------------------------------------------
; See if we need to do the top half of the triangle
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[TopHeight]          ;
  Cmp  edx,0                    ;
  Jle  @@SkipTopTriangleRender  ;
                                ;
;----------------------------------------------------------------------------
; Calculate the short side data
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [ebx].ScrX     ;
  fsub dword ptr [eax].ScrX     ; (VB->Scr.x - VA->Scr.x)
  fmul dword ptr [OneOverTopDY] ; (VB->Scr.x - VA->Scr.x) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDShortDX]  ;  store fixed point
                                ;
  fild  dword ptr [CeilY1]      ;
  fsub  dword ptr [eax].ScrY    ;
  fmulp st(1),st(0)             ;
  fadd  dword ptr [eax].ScrX    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDShortX]   ;  store fixed point
                                ;
;----------------------------------------------------------------------------
; Compare deltas to see which order we draw the polygon
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[FXDLongDX]          ;
  Cmp  edx,[FXDShortDX]         ; if (longdx < shortdx)
                                ;
  Jge @@LongSideRight           ;
                                ;
  INNER_LOOP FXDLong, FXDShort, TopHeight, FXDSide1
                                ;
  Jmp @@SkipTopTriangleRender   ;
 @@LongSideRight:               ;
                                ;
  INNER_LOOP FXDShort, FXDLong, TopHeight, FXDSide1
                                ;
 @@SkipTopTriangleRender:       ;
                                ;
;----------------------------------------------------------------------------
; See if we need to do the bottom half of the triangle
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[BottomHeight]       ;
  Cmp  edx,0                    ;
  Jle  @@SkipBottomTriangleRender;
                                ;
  Mov  eax,[VertPtrA]           ;  re-grab the vertex pointers
  Mov  ebx,[VertPtrB]           ;
  Mov  ecx,[VertPtrC]           ;
                                ;
;----------------------------------------------------------------------------
; Calculate the short side data for bottom
;----------------------------------------------------------------------------
                                ;
  fld  dword ptr [ecx].ScrX     ;
  fsub dword ptr [ebx].ScrX     ; (VC->Scr.x - VB->Scr.x)
  fmul dword ptr [OneOverBotDY] ; (VC->Scr.x - VB->Scr.x) * OneOverDY
                                ;
  fld   st(0)                   ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDShortDX]  ;  store fixed point
                                ;
  fild  dword ptr [CeilY2]      ;
  fsub  dword ptr [ebx].ScrY    ;
  fmulp st(1),st(0)             ;
  fadd  dword ptr [ebx].ScrX    ;
  fmul  dword ptr [_tmp_fixone] ;
  fistp dword ptr [FXDShortX]   ;  store fixed point
                                ;
;----------------------------------------------------------------------------
; Compare deltas to see which order we draw the polygon for bottom
;----------------------------------------------------------------------------
                                ;
  Mov  edx,[FXDLongDX]          ;
  Cmp  edx,[FXDShortDX]         ; if (longdx < shortdx)
                                ;
  Jge @@RasterLongSideRight     ;
                                ;
  INNER_LOOP FXDShort, FXDLong, BottomHeight, FXDSide2
                                ;
  Jmp @@ExitPolygonFiller       ;
 @@RasterLongSideRight:         ;
                                ;
  INNER_LOOP FXDLong, FXDShort, BottomHeight, FXDSide1
                                ;
 @@SkipBottomTriangleRender:    ;
 @@ExitPolygonFiller:           ;
