/*

                       Inertia Realtime 3D Rendering Engine
            Copyright (c) 1997, Alex Chalfin. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED

*/

#include "pool.h"

typedef struct VertexNode
{
  Vertex VList[POOL_NODE_COUNT];
  struct VertexNode *Next;
} VertexNode;

typedef struct PolygonNode
{
  Triangle PList[POOL_NODE_COUNT];
  struct  PolygonNode *Next;
} PolygonNode;

struct
{
  int        Count;
  int        Allocated;
  VertexNode *Traverse;
  VertexNode *VertexList;
} VertexPool;

struct
{
  int         Count;
  int         Allocated;
  PolygonNode *Traverse;
  PolygonNode *PolygonList;
} PolygonPool;


Vertex *GrabVertexFromPool()
{
  VertexNode *TempPtr;

  if ((VertexPool.Count + 1) >= VertexPool.Allocated)
  {
    TempPtr = (VertexNode *)malloc(sizeof(VertexNode));

    if (TempPtr == NULL)
      return(NULL);

    TempPtr->Next = NULL;
    if (VertexPool.VertexList == NULL)
    {
      VertexPool.VertexList = TempPtr;
      VertexPool.Traverse   = TempPtr;
    }
    else
    {
      VertexPool.Traverse->Next = TempPtr;
      VertexPool.Traverse       = VertexPool.Traverse->Next;
    }

    VertexPool.Allocated += POOL_NODE_COUNT;
  }
  else
  {
    if (((VertexPool.Count + 1) & POOL_NODE_MASK) < ((VertexPool.Count) & POOL_NODE_MASK))
      VertexPool.Traverse = VertexPool.Traverse->Next;
  }

  return(&(VertexPool.Traverse->VList[(VertexPool.Count++) & POOL_NODE_MASK]));
}

Triangle *GrabPolygonFromPool()
{
  PolygonNode *TempPtr;

  if ((PolygonPool.Count + 1) >= PolygonPool.Allocated)
  {
    TempPtr = (PolygonNode *)malloc(sizeof(PolygonNode));

    if (TempPtr == NULL)
      return(NULL);

    TempPtr->Next = NULL;
    if (PolygonPool.PolygonList == NULL)
    {
      PolygonPool.PolygonList = TempPtr;
      PolygonPool.Traverse   = TempPtr;
    }
    else
    {
      PolygonPool.Traverse->Next = TempPtr;
      PolygonPool.Traverse       = PolygonPool.Traverse->Next;
    }

    PolygonPool.Allocated += POOL_NODE_COUNT;
  }
  else
  {
    if (((PolygonPool.Count + 1) & POOL_NODE_MASK) < ((PolygonPool.Count) & POOL_NODE_MASK))
      PolygonPool.Traverse = PolygonPool.Traverse->Next;
  }

  return(&(PolygonPool.Traverse->PList[(PolygonPool.Count++) & POOL_NODE_MASK]));
}

void ResetVertexPool()
{
  VertexPool.Count    = 0;
  VertexPool.Traverse = VertexPool.VertexList;
}

void ResetPolygonPool()
{
  PolygonPool.Count    = 0;
  PolygonPool.Traverse = PolygonPool.PolygonList;
}

void InitVertexPool()
{
  VertexPool.Count      = 0;
  VertexPool.Allocated  = 0;
  VertexPool.VertexList = NULL;
  VertexPool.Traverse   = NULL;
}

void InitPolygonPool()
{
  PolygonPool.Count       = 0;
  PolygonPool.Allocated   = 0;
  PolygonPool.Traverse    = NULL;
  PolygonPool.PolygonList = NULL;
}

void FreePolygonPool()
{
  PolygonNode *TempPtr;

  PolygonPool.Allocated = 0;
  PolygonPool.Count     = 0;

  PolygonPool.Traverse = PolygonPool.PolygonList;

  while (PolygonPool.Traverse != NULL)
  {
    TempPtr = PolygonPool.Traverse;
    PolygonPool.Traverse = PolygonPool.Traverse->Next;
    free(TempPtr);
  }

  PolygonPool.PolygonList = NULL;
}

void FreeVertexPool()
{
  VertexNode *TempPtr;

  VertexPool.Allocated = 0;
  VertexPool.Count     = 0;

  VertexPool.Traverse = VertexPool.VertexList;

  while (VertexPool.Traverse != NULL)
  {
    TempPtr = VertexPool.Traverse;
    VertexPool.Traverse = VertexPool.Traverse->Next;
    free(TempPtr);
  }

  VertexPool.VertexList = NULL;
}
